package letsencrypt

  1. Overview
  2. Docs

Module Letsencrypt.ClientSource

ACME Client.

This module provides client commands. Note: right now this module implements only the strict necessary in order to register an account, solve http-01 challenges provided by the CA, and fetch the certificate. This means that you will be able to maintain your server with this.

Sourcetype t
Sourcetype challenge =
  1. | DNS
  2. | HTTP
  3. | ALPN

Scheduler monad

Sourcemodule type S = sig ... end

HTTP client interface

Sourcemodule type Client = sig ... end
Sourcemodule Solver (S : S) : sig ... end
Sourcemodule Make (S : S) (C : Client with type 'a t = 'a S.t) : sig ... end